"""
A command line interface to the qcfractal server.
"""

import argparse
import json
import logging
import os
import signal
from enum import Enum
from math import ceil
from typing import List, Optional, Union

import tornado.log
import yaml
from pydantic import Field, validator

import qcengine as qcng
import qcfractal

from ..interface.models import AutodocBaseSettings, ProtoModel
from . import cli_utils

__all__ = ["main"]

QCA_RESOURCE_STRING = "--resources process=1"

logger = logging.getLogger("qcfractal.cli")


class SettingsCommonConfig:
    env_prefix = "QCA_"
    case_insensitive = True
    extra = "forbid"


class AdapterEnum(str, Enum):
    dask = "dask"
    pool = "pool"
    parsl = "parsl"


class CommonManagerSettings(AutodocBaseSettings):
    """
    The Common settings are the settings most users will need to adjust regularly to control the nature of
    task execution and the hardware under which tasks are executed on. This block is often unique to each deployment,
    user, and manager and will be the most commonly updated options, even as config files are copied and reused, and
    even on the same platform/cluster.

    """

    adapter: AdapterEnum = Field(
        AdapterEnum.pool, description="Which type of Distributed adapter to run tasks through."
    )
    tasks_per_worker: int = Field(
        1,
        description="Number of concurrent tasks to run *per Worker* which is executed. Total number of concurrent "
        "tasks is this value times max_workers, assuming the hardware is available. With the "
        "pool adapter, and/or if max_workers=1, tasks_per_worker *is* the number of concurrent tasks.",
    )
    cores_per_worker: int = Field(
        qcng.config.get_global("ncores"),
        description="""
        Number of cores to be consumed by the Worker and distributed over the tasks_per_worker. These 
        cores are divided evenly, so it is recommended that quotient of cores_per_worker/tasks_per_worker 
        be a whole number else the core distribution is left up to the logic of the adapter. The default
        value is read from the number of detected cores on the system you are executing on.
        
        In the case of node-parallel tasks, this number means the number of cores per node.
        """,
        gt=0,
    )
    memory_per_worker: float = Field(
        qcng.config.get_global("memory"),
        description="Amount of memory (in GB) to be consumed and distributed over the tasks_per_worker. This memory is "
        "divided evenly, but is ultimately at the control of the adapter. Engine will only allow each of "
        "its calls to consume memory_per_worker/tasks_per_worker of memory. Total memory consumed by this "
        "manager at any one time is this value times max_workers. The default value is read "
        "from the amount of memory detected on the system you are executing on.",
        gt=0,
    )
    max_workers: int = Field(
        1,
        description="The maximum number of Workers which are allowed to be run at the same time. The total number of "
        "concurrent tasks will maximize at this quantity times tasks_per_worker."
        "The total number "
        "of Jobs on a cluster which will be started is equal to this parameter in most cases, and should "
        "be assumed 1 Worker per Job. Any exceptions to this will be documented. "
        "In node exclusive mode this is equivalent to the maximum number of nodes which you will consume. "
        "This must be a positive, non zero integer.",
        gt=0,
    )
    retries: int = Field(
        2,
        description="Number of retries that QCEngine will attempt for RandomErrors detected when running "
        "its computations. After this many attempts (or on any other type of error), the "
        "error will be raised.",
        ge=0,
    )
    scratch_directory: Optional[str] = Field(None, description="Scratch directory for Engine execution jobs.")
    verbose: bool = Field(
        False,
        description="Turn on verbose mode or not. In verbose mode, all messages from DEBUG level and up are shown, "
        "otherwise, defaults are all used for any logger.",
    )
    nodes_per_job: int = Field(
        1, description="The number of nodes to request per job. Only used by the Parsl adapter at present", gt=0
    )
    nodes_per_task: int = Field(
        1, description="The number of nodes to use for each tasks. Only relevant for node-parallel executables.", gt=0
    )
    cores_per_rank: int = Field(
        1,
        description="The number of cores per MPI rank for MPI-parallel applications. Only relevant for node-parallel"
        " codes and the most relevant to codes that with hybrid MPI+OpenMP parallelism (e.g., NWChem).",
    )

    class Config(SettingsCommonConfig):
        pass


class FractalServerSettings(AutodocBaseSettings):
    """
    Settings pertaining to the Fractal Server you wish to pull tasks from and push completed tasks to. Each manager
    supports exactly 1 Fractal Server to be in communication with, and exactly 1 user on that Fractal Server. These
    can be changed, but only once the Manager is shutdown and the settings changed. Multiple Managers however can be
    started in parallel with each other, but must be done as separate calls to the CLI.

    Caution: The password here is written in plain text, so it is up to the owner/writer of the configuration file
    to ensure its security.
    """

    fractal_uri: str = Field("localhost:7777", description="Full URI to the Fractal Server you want to connect to")
    username: Optional[str] = Field(
        None,
        description="Username to connect to the Fractal Server with. When not provided, a connection is attempted "
        "as a guest user, which in most default Servers will be unable to return results.",
    )
    password: Optional[str] = Field(
        None, description="Password to authenticate to the Fractal Server with (alongside the `username`)"
    )
    verify: Optional[bool] = Field(None, description="Use Server-side generated SSL certification or not.")

    class Config(SettingsCommonConfig):
        pass


class QueueManagerSettings(AutodocBaseSettings):
    """
    Fractal Queue Manger settings. These are options which control the setup and execution of the Fractal Manager
    itself.
    """

    manager_name: str = Field(
        "unlabeled",
        description="Name of this scheduler to present to the Fractal Server. Descriptive names help the server "
        "identify the manager resource and assists with debugging.",
    )
    queue_tag: Optional[Union[str, List[str]]] = Field(
        None,
        description="Only pull tasks from the Fractal Server with this tag. If not set (None/null), then pull untagged "
        "tasks, which should be the majority of tasks. This option should only be used when you want to "
        "pull very specific tasks which you know have been tagged as such on the server. If the server has "
        "no tasks with this tag, no tasks will be pulled (and no error is raised because this is intended "
        "behavior). If multiple tags are provided, tasks will be pulled (but not necessarily executed) in order of the "
        "tags.",
    )
    log_file_prefix: Optional[str] = Field(
        None,
        description="Full path to save a log file to, including the filename. If not provided, information will still "
        "be reported to terminal, but not saved. When set, logger information is sent both to this file "
        "and the terminal.",
    )
    update_frequency: float = Field(
        30,
        description="Time between heartbeats/update checks between this Manager and the Fractal Server. The lower this "
        "value, the shorter the intervals. If you have an unreliable network connection, consider "
        "increasing this time as repeated, consecutive network failures will cause the Manager to shut "
        "itself down to maintain integrity between it and the Fractal Server. Units of seconds",
        gt=0,
    )
    test: bool = Field(
        False,
        description="Turn on testing mode for this Manager. The Manager will not connect to any Fractal Server, and "
        "instead submits netsts worth trial tasks per quantum chemistry program it finds. These tasks are "
        "generated locally and do not need a running Fractal Server to work. Helpful for ensuring the "
        "Manager is configured correctly and the quantum chemistry codes are operating as expected.",
    )
    ntests: int = Field(
        5,
        description="Number of tests to run if the `test` flag is set to True. Total number of tests will be this "
        "number times the number of found quantum chemistry programs. Does nothing if `test` is False."
        "If set to 0, then this submits no tests, but it will run through the setup and client "
        "initialization.",
        gt=-1,
    )
    max_queued_tasks: Optional[int] = Field(
        None,
        description="Generally should not be set. Number of tasks to pull from the Fractal Server to keep locally at "
        "all times. If `None`, this is automatically computed as "
        "`ceil(common.tasks_per_worker*common.max_workers*2.0) + 1`. As tasks are completed, the "
        "local pool is filled back up to this value. These tasks will all attempt to be run concurrently, "
        "but concurrent tasks are limited by number of cluster jobs and tasks per job. Pulling too many of "
        "these can result in under-utilized managers from other sites and result in less FIFO returns. As "
        "such it is recommended not to touch this setting in general as you will be given enough tasks to "
        "fill your maximum throughput with a buffer (assuming the queue has them).",
        gt=0,
    )


class SchedulerEnum(str, Enum):
    slurm = "slurm"
    pbs = "pbs"
    sge = "sge"
    moab = "moab"
    lsf = "lsf"
    cobalt = "cobalt"


class AdaptiveCluster(str, Enum):
    static = "static"
    adaptive = "adaptive"


class ClusterSettings(AutodocBaseSettings):
    """
    Settings tied to the cluster you are running on. These settings are mostly tied to the nature of the cluster
    jobs you are submitting, separate from the nature of the compute tasks you will be running within them. As such,
    the options here are things like wall time (per job), which Scheduler your cluster has (like PBS or SLURM),
    etc. No additional options are allowed here.
    """

    node_exclusivity: bool = Field(
        False,
        description="Run your cluster jobs in node-exclusivity mode. This option may not be available to all scheduler "
        "types and thus may not do anything. Related to this, the flags we have found for this option "
        "may not be correct for your scheduler and thus might throw an error. You can always add the "
        "correct flag/parameters to the `scheduler_options` parameter and leave this as False if you "
        "find it gives you problems.",
    )
    scheduler: SchedulerEnum = Field(
        None,
        description="Option of which Scheduler/Queuing system your cluster uses. Note: not all scheduler options are "
        "available with every adapter.",
    )
    scheduler_options: List[str] = Field(
        [],
        description="Additional options which are fed into the header files for your submitted jobs to your cluster's "
        "Scheduler/Queuing system. The directives are automatically filled in, so if you want to set "
        "something like '#PBS -n something', you would instead just do '-n something'. Each directive "
        "should be a separate string entry in the list. No validation is done on this with respect to "
        "valid directives so it is on the user to know what they need to set.",
    )
    task_startup_commands: List[str] = Field(
        [],
        description="Additional commands to be run before starting the Workers and the task distribution. This can "
        "include commands needed to start things like conda environments or setting environment variables "
        "before executing the Workers. These commands are executed first before any of the distributed "
        "commands run and are added to the batch scripts as individual commands per entry, verbatim.",
    )
    walltime: str = Field(
        "06:00:00",
        description="Wall clock time of each cluster job started. Presented as a string in HH:MM:SS form, but your "
        "cluster may have a different structural syntax. This number should be set high as there should "
        "be a number of Fractal tasks which are run for each submitted cluster job. Ideally, the job "
        "will start, the Worker will land, and the Worker will crunch through as many tasks as it can; "
        "meaning the job which has a Worker in it must continue existing to minimize time spend "
        "redeploying Workers.",
    )
    adaptive: AdaptiveCluster = Field(
        AdaptiveCluster.adaptive,
        description="Whether or not to use adaptive scaling of Workers or not. If set to 'static', a fixed number of "
        "Workers will be started (and likely *NOT* restarted when the wall clock is reached). When set to "
        "'adaptive' (the default), the distributed engine will try to adaptively scale the number of "
        "Workers based on tasks in the queue. This is str instead of bool type variable in case more "
        "complex adaptivity options are added in the future.",
    )

    class Config(SettingsCommonConfig):
        pass

    @validator("scheduler", "adaptive", pre=True)
    def _lcase(cls, v):
        if v:
            v = v.lower()
        return v


class SettingsBlocker(AutodocBaseSettings):
    """Helper class to auto block certain entries, overwrite hidden methods to access"""

    _forbidden_set = set()
    _forbidden_name = "SettingsBlocker"

    def __init__(self, **kwargs):
        """
        Enforce that the keys we are going to set remain untouched. Blocks certain keywords for the classes
        they will be fed into, not whatever Fractal is using as keywords.
        """
        bad_set = set(kwargs.keys()) & self._forbidden_set
        if bad_set:
            raise KeyError(
                "The following items were set as part of {}, however, "
                "there are other config items which control these in more generic "
                "settings locations: {}".format(self._forbidden_name, bad_set)
            )
        super().__init__(**kwargs)

    class Config(SettingsCommonConfig):
        # This overwrites the base config to allow other keywords to be fed in
        extra = "allow"


class DaskQueueSettings(SettingsBlocker):
    """
    Settings for the Dask Cluster class. Values set here are passed directly into the Cluster objects based on the
    `cluster.scheduler` settings. Although many values are set automatically from other settings, there are
    some additional values such as `interface` and `extra` which are passed through to the constructor.

    Valid values for this field are functions of your cluster.scheduler and no linting is done ahead of trying to pass
    these to Dask.

    NOTE: The parameters listed here are a special exception for additional features Fractal has engineered or
    options which should be considered for some of the edge cases we have discovered. If you try to set a value
    which is derived from other options in the YAML file, an error is raised and you are told exactly which one is
    forbidden.

    Please see the docs for the provider for more information.
    """

    interface: Optional[str] = Field(
        None,
        description="Name of the network adapter to use as communication between the head node and the compute node."
        "There are oddities of this when the head node and compute node use different ethernet adapter "
        "names and we have not figured out exactly which combination is needed between this and the "
        "poorly documented `ip` keyword which appears to be for Workers, but not the Client.",
    )
    extra: Optional[List[str]] = Field(
        None,
        description="Additional flags which are fed into the Dask Worker CLI startup, can be used to overwrite "
        "pre-configured options. Do not use unless you know exactly which flags to use.",
    )
    lsf_units: Optional[str] = Field(
        None,
        description="Unit system for an LSF cluster limits (e.g. MB, GB, TB). If not set, the units are "
        "are attempted to be set from the `lsf.conf` file in the default locations. This does nothing "
        "if the cluster is not LSF",
    )
    _forbidden_set = {"name", "cores", "memory", "processes", "walltime", "env_extra", "qca_resource_string"}
    _forbidden_name = "dask_jobqueue"


class ParslExecutorSettings(SettingsBlocker):
    """
    Settings for the Parsl Executor class. This serves as the primary mechanism for distributing Workers to jobs.
    In most cases, you will not need to set any of these options, as several options are automatically inferred
    from other settings. Any option set here is passed through to the HighThroughputExecutor class of Parsl.

    https://parsl.readthedocs.io/en/latest/stubs/parsl.executors.HighThroughputExecutor.html

    NOTE: The parameters listed here are a special exception for additional features Fractal has engineered or
    options which should be considered for some of the edge cases we have discovered. If you try to set a value
    which is derived from other options in the YAML file, an error is raised and you are told exactly which one is
    forbidden.

    """

    address: Optional[str] = Field(
        None,
        description="This only needs to be set in conditional cases when the head node and compute nodes use a "
        "differently named ethernet adapter.\n\n"
        "An address to connect to the main Parsl process which is reachable from the network in which "
        "Workers will be running. This can be either a hostname as returned by hostname or an IP address. "
        "Most login nodes on clusters have several network interfaces available, only some of which can be "
        "reached from the compute nodes. Some trial and error might be necessary to identify what "
        "addresses are reachable from compute nodes.",
    )
    _forbidden_set = {"label", "provider", "cores_per_worker", "max_workers"}
    _forbidden_name = "the parsl executor"


class ParslLauncherSettings(AutodocBaseSettings):
    """
    Set the Launcher in a Parsl Provider, and its options, if not set, the defaults are used.

    This is a rare use case where the ``launcher`` key of the Provider is needed to be set. Since it must be a class
    first, you will need to specify the ``launcher_type`` options which is interpreted as the Class Name of the
    Launcher to load and pass the rest of the options set here into it. Any unset key will just be left as defaults.
    It is up to the user to consult the Parsl Docs for their desired Launcher's options and what they do.

    The known launchers below are case-insensitive,
    but if new launchers come out (or you are using a custom/developmental build of Parsl), then you can pass your
    own Launcher in verbatim, with case sensitivity, and the Queue Manager will try to load it.

    Known Launchers:
        - ``SimpleLauncher``: https://parsl.readthedocs.io/en/latest/stubs/parsl.launchers.SimpleLauncher.html
        - ``SingleNodeLauncher``: https://parsl.readthedocs.io/en/latest/stubs/parsl.launchers.SingleNodeLauncher.html
        - ``SrunLauncher``: https://parsl.readthedocs.io/en/latest/stubs/parsl.launchers.SrunLauncher.html
        - ``AprunLauncher``: https://parsl.readthedocs.io/en/latest/stubs/parsl.launchers.AprunLauncher.html
        - ``SrunMPILauncher``: https://parsl.readthedocs.io/en/latest/stubs/parsl.launchers.SrunMPILauncher.html
        - ``GnuParallelLauncher``: https://parsl.readthedocs.io/en/latest/stubs/parsl.launchers.GnuParallelLauncher.html
        - ``MpiExecLauncher`` : https://parsl.readthedocs.io/en/latest/stubs/parsl.launchers.MpiExecLauncher.html
    """

    launcher_class: str = Field(
        ...,
        description="Class of Launcher to use. This is a setting unique to QCArchive which is then used to pass onto "
        "the Provider's ``launcher`` setting and the remaining keys are passed to that Launcher's options.",
    )

    def _get_launcher(self, launcher_base: str) -> "Launcher":
        launcher_lower = launcher_base.lower()
        launcher_map = {
            "simplelauncher": "SimpleLauncher",
            "singlenodelauncher": "SingleNodeLauncher",
            "srunlauncher": "SrunLauncher",
            "aprunlauncher": "AprunLauncher",
            "srunmpiLauncher": "SrunMPILauncher",
            "gnuparallellauncher": "GnuParallelLauncher",
            "mpiexeclauncher": "MpiExecLauncher",
        }
        launcher_string = launcher_map[launcher_lower] if launcher_lower in launcher_map else launcher_base

        try:
            launcher_load = cli_utils.import_module("parsl.launchers", package=launcher_string)
            launcher = getattr(launcher_load, launcher_string)
        except ImportError:
            raise ImportError(
                f"Could not import Parsl Launcher: {launcher_base}. Please make sure you have Parsl "
                f"installed and are requesting one of the launchers within the package."
            )
        return launcher

    def build_launcher(self):
        """Import and load the desired launcher"""
        launcher = self._get_launcher(self.launcher_class)
        return launcher(**self.dict(exclude={"launcher_class"}))

    class Config(SettingsCommonConfig):
        extra = "allow"


class ParslProviderSettings(SettingsBlocker):
    """
    Settings for the Parsl Provider class. Valid values for this field depend on your choice of  cluster.scheduler and
    are defined in `the Parsl docs for the providers
    <https://parsl.readthedocs.io/en/stable/userguide/execution.html#execution-providers>`_ with some minor exceptions.
    The initializer function for the Parsl settings will indicate which

    NOTE: The parameters listed here are a special exception for additional features Fractal has engineered or
    options which should be considered for some of the edge cases we have discovered. If you try to set a value
    which is derived from other options in the YAML file, an error is raised and you are told exactly which one is
    forbidden.

    SLURM: https://parsl.readthedocs.io/en/latest/stubs/parsl.providers.SlurmProvider.html
    PBS/Torque/Moab: https://parsl.readthedocs.io/en/latest/stubs/parsl.providers.TorqueProvider.html
    SGE (Sun GridEngine): https://parsl.readthedocs.io/en/latest/stubs/parsl.providers.GridEngineProvider.html

    """

    def __init__(self, **kwargs):
        if "max_blocks" in kwargs:
            raise ValueError("``max_blocks`` is set based on ``common.max_workers`` " "and ``common.nodes_per_job``")
        super().__init__(**kwargs)

    partition: str = Field(
        None,
        description="The name of the cluster.scheduler partition being submitted to. Behavior, valid values, and even"
        "its validity as a set variable are a function of what type of queue scheduler your specific "
        "cluster has (e.g. this variable should NOT be present for PBS clusters). "
        "Check with your Sys. Admins and/or your cluster documentation.",
    )
    launcher: ParslLauncherSettings = Field(
        None,
        description="The Parsl Launcher to use with your Provider. If left to ``None``, defaults are assumed (check "
        "the Provider's defaults), otherwise this should be a dictionary requiring the option "
        "``launcher_class`` as a str to specify which Launcher class to load, and the remaining settings "
        "will be passed on to the Launcher's constructor.",
    )
    _forbidden_set = {"worker_init", "scheduler_options", "wall_time", "nodes_per_block"}
    _forbidden_name = "parsl's provider"


class ParslQueueSettings(AutodocBaseSettings):
    """
    The Parsl-specific configurations used with the `common.adapter = parsl` setting. The parsl config is broken up into
    a top level `Config` class, an `Executor` sub-class, and a `Provider` sub-class of the `Executor`.
    Config -> Executor -> Provider. Each of these have their own options, and extra values fed into the
    ParslQueueSettings are fed to the `Config` level.

    It requires both `executor` and `provider` settings, but will default fill them in and often does not need
    any further configuration which is handled by other settings in the config file.
    """

    executor: ParslExecutorSettings = ParslExecutorSettings()
    provider: ParslProviderSettings = ParslProviderSettings()

    class Config(SettingsCommonConfig):
        extra = "allow"


class ManagerSettings(ProtoModel):
    """
    The config file for setting up a QCFractal Manager, all sub fields of this model are at equal top-level of the
    YAML file. No additional top-level fields are permitted, but sub-fields may have their own additions.

    Not all fields are required and many will depend on the cluster you are running, and the adapter you choose
    to run on.
    """

    common: CommonManagerSettings = CommonManagerSettings()
    server: FractalServerSettings = FractalServerSettings()
    manager: QueueManagerSettings = QueueManagerSettings()
    cluster: Optional[ClusterSettings] = ClusterSettings()
    dask: Optional[DaskQueueSettings] = DaskQueueSettings()
    parsl: Optional[ParslQueueSettings] = ParslQueueSettings()

    class Config(ProtoModel.Config):
        extra = "forbid"


def parse_args():
    parser = argparse.ArgumentParser(
        description="A CLI for a QCFractal QueueManager with a ProcessPoolExecutor, Dask, or Parsl backend. "
        "The Dask and Parsl backends *requires* a config file due to the complexity of its setup. If a config "
        "file is specified, the remaining options serve as CLI overwrites of the config."
    )
    parser.add_argument("--version", action="version", version=f"{qcfractal.__version__}")

    parser.add_argument("--config-file", type=str, default=None)

    # Common settings
    common = parser.add_argument_group("Common Adapter Settings")
    common.add_argument(
        "--adapter", type=str, help="The backend adapter to use, currently only {'dask', 'parsl', 'pool'} are valid."
    )
    common.add_argument(
        "--tasks-per-worker",
        type=int,
        help="The number of simultaneous tasks for the executor to run, resources will be divided evenly.",
    )
    common.add_argument("--cores-per-worker", type=int, help="The number of process for each executor's Workers")
    common.add_argument("--memory-per-worker", type=int, help="The total amount of memory on the system in GB")
    common.add_argument("--scratch-directory", type=str, help="Scratch directory location")
    common.add_argument("--retries", type=int, help="Number of RandomError retries per task before failing the task")
    common.add_argument("-v", "--verbose", action="store_true", help="Increase verbosity of the logger.")

    # FractalClient options
    server = parser.add_argument_group("FractalServer connection settings")
    server.add_argument("--fractal-uri", type=str, help="FractalServer location to pull from")
    server.add_argument("-u", "--username", type=str, help="FractalServer username")
    server.add_argument("-p", "--password", type=str, help="FractalServer password")
    server.add_argument(
        "--verify",
        type=str,
        help="Do verify the SSL certificate, leave off (unset) for servers with custom SSL certificates.",
    )

    # QueueManager options
    manager = parser.add_argument_group("QueueManager settings")
    manager.add_argument("--manager-name", type=str, help="The name of the manager to start")
    manager.add_argument("--queue-tag", type=str, help="The queue tag to pull from")
    manager.add_argument("--log-file-prefix", type=str, help="The path prefix of the logfile to write to.")
    manager.add_argument("--update-frequency", type=int, help="The frequency in seconds to check for complete tasks.")
    manager.add_argument(
        "--max-queued-tasks",
        type=int,
        help="Maximum number of tasks to hold at any given time. " "Generally should not be set.",
    )

    # Additional args
    optional = parser.add_argument_group("Optional Settings")
    optional.add_argument("--test", action="store_true", help="Boot and run a short test suite to validate setup")
    optional.add_argument(
        "--ntests", type=int, help="How many tests per found program to run, does nothing without --test set"
    )
    optional.add_argument(
        "--schema",
        action="store_true",
        help="Display the current Schema (Pydantic) for the YAML "
        "config file and exit. This will always show the "
        "most up-to-date schema. It will be presented in a "
        "JSON-like format.",
    )
    optional.add_argument(
        "--skeleton",
        "--skel",
        type=str,
        const="manager_config.yaml",
        default=None,
        action="store",
        nargs="?",
        help="Create a skeleton/example YAML config file at the specified path. This does not start "
        "the manager and instead creates a skeleton based on all the options specified.",
    )

    # Move into nested namespace
    args = vars(parser.parse_args())

    def _build_subset(args, keys):
        ret = {}
        for k in keys:
            v = args[k]

            if v is None:
                continue

            ret[k] = v
        return ret

    # Stupid we cannot inspect groups
    data = {
        "common": _build_subset(
            args,
            {
                "adapter",
                "tasks_per_worker",
                "cores_per_worker",
                "memory_per_worker",
                "scratch_directory",
                "retries",
                "verbose",
            },
        ),
        "server": _build_subset(args, {"fractal_uri", "password", "username", "verify"}),
        "manager": _build_subset(
            args,
            {"max_queued_tasks", "manager_name", "queue_tag", "log_file_prefix", "update_frequency", "test", "ntests"},
        ),
        # This set is for this script only, items here should not be passed to the ManagerSettings nor any other
        # classes
        "debug": _build_subset(args, {"schema", "skeleton"}),
    }  # yapf: disable

    if args["config_file"] is not None:
        config_data = cli_utils.read_config_file(args["config_file"])
        for name, subparser in [("common", common), ("server", server), ("manager", manager)]:
            if name not in config_data:
                continue

            data[name] = cli_utils.argparse_config_merge(subparser, data[name], config_data[name], check=False)

        for name in ["cluster", "dask", "parsl"]:
            if name in config_data:
                data[name] = config_data[name]
                if data[name] is None:
                    # Handle edge case where None provided here is explicitly treated as
                    # "do not parse" by Pydantic (intended behavior) instead of the default empty dict
                    # being used instead. This only happens when a user sets in the YAML file
                    # the top level header and nothing below it.
                    data[name] = {}

    return data


def main(args=None):

    # Grab CLI args if not present
    if args is None:
        args = parse_args()
    exit_callbacks = []

    try:
        if args["debug"]["schema"]:
            print(ManagerSettings.schema_json(indent=2))
            return  # We're done, exit normally
    except KeyError:
        pass  # Don't worry if schema isn't in the list
    finally:
        debug_args = args.pop("debug", {})  # Ensure the debug key is not present

    # Construct object
    settings = ManagerSettings(**args)

    # Handle Skeleton Generation
    if debug_args.get("skeleton", None):

        class IndentListDumper(yaml.Dumper):
            """
            Internal yaml Dumper to make lists indent in the output YAML

            Buried inside this since its only used in "skeleton," once, and then exits. Does not need to be imported
            anywhere else or accessed somehow

            Based on response:
            https://stackoverflow.com/questions/25108581/python-yaml-dump-bad-indentation/39681672#39681672
            """

            def increase_indent(self, flow=False, indentless=False):
                return super(IndentListDumper, self).increase_indent(flow, False)

        skel_path = os.path.expanduser(debug_args["skeleton"])
        with open(skel_path, "w") as skel:
            # cast to
            data = yaml.dump(json.loads(settings.json()), Dumper=IndentListDumper, default_flow_style=False)
            skel.write(data)
            print(
                f"Skeleton Queue Manager YAML file written to {skel_path}\n"
                f"Run: `qcfractal-manager --config-file={skel_path}` to start a manager with this configuration."
            )
            return

    logger_map = {AdapterEnum.pool: "", AdapterEnum.dask: "dask_jobqueue.core", AdapterEnum.parsl: "parsl"}
    if settings.common.verbose:
        adapter_logger = logging.getLogger(logger_map[settings.common.adapter])
        adapter_logger.setLevel("DEBUG")
        logger.setLevel("DEBUG")

    if settings.manager.log_file_prefix is not None:
        tornado.options.options["log_file_prefix"] = settings.manager.log_file_prefix
        # Clones the log to the output
        tornado.options.options["log_to_stderr"] = True
    tornado.log.enable_pretty_logging()

    if settings.manager.test:
        # Test this manager, no client needed
        client = None
    else:
        # Connect to a specified fractal server
        client = qcfractal.interface.FractalClient(
            address=settings.server.fractal_uri, **settings.server.dict(skip_defaults=True, exclude={"fractal_uri"})
        )

    # Figure out per-task data
    node_parallel_tasks = settings.common.nodes_per_task > 1  # Whether tasks are node-parallel
    if node_parallel_tasks:
        supported_adapters = ["parsl"]
        if settings.common.adapter not in supported_adapters:
            raise ValueError("Node-parallel jobs are only supported with {} adapters".format(supported_adapters))
        # Node-parallel tasks use all cores on a worker
        cores_per_task = settings.common.cores_per_worker
        memory_per_task = settings.common.memory_per_worker
        if settings.common.tasks_per_worker > 1:
            raise ValueError(">1 task per node and >1 node per tasks are mutually-exclusive")
    else:
        cores_per_task = settings.common.cores_per_worker // settings.common.tasks_per_worker
        memory_per_task = settings.common.memory_per_worker / settings.common.tasks_per_worker
    if cores_per_task < 1:
        raise ValueError("Cores per task must be larger than one!")

    if settings.common.adapter == "pool":
        from concurrent.futures import ProcessPoolExecutor

        # TODO: Replace with passing via mp_context to ProcessPoolExecutor
        # when python 3.6 is dead and buried
        from multiprocessing import set_start_method

        set_start_method("spawn")

        # Error if the number of nodes per jobs is more than 1
        if settings.common.nodes_per_job > 1:
            raise ValueError("Pool adapters only run on a single local node")
        queue_client = ProcessPoolExecutor(max_workers=settings.common.tasks_per_worker)

    elif settings.common.adapter == "dask":

        dask_settings = settings.dask.dict(skip_defaults=True)
        # Checks
        if "extra" not in dask_settings:
            dask_settings["extra"] = []
        if QCA_RESOURCE_STRING not in dask_settings["extra"]:
            dask_settings["extra"].append(QCA_RESOURCE_STRING)
        # Scheduler opts
        scheduler_opts = settings.cluster.scheduler_options.copy()

        # Error if the number of nodes per jobs is more than 1
        if settings.common.nodes_per_job > 1:
            raise NotImplementedError("Support for >1 node per job is not yet supported by QCFractal + Dask")
            # TODO (wardlt): Implement multinode jobs in Dask

        _cluster_loaders = {
            "slurm": "SLURMCluster",
            "pbs": "PBSCluster",
            "moab": "MoabCluster",
            "sge": "SGECluster",
            "lsf": "LSFCluster",
        }
        dask_exclusivity_map = {
            "slurm": "--exclusive",
            "pbs": "-n",
            "moab": "-n",  # Less sure about this one
            "sge": "-l exclusive=true",
            "lsf": "-x",
        }
        if settings.cluster.node_exclusivity and dask_exclusivity_map[settings.cluster.scheduler] not in scheduler_opts:
            scheduler_opts.append(dask_exclusivity_map[settings.cluster.scheduler])

        # Create one construct to quickly merge dicts with a final check
        dask_construct = {
            "name": "QCFractal_Dask_Compute_Executor",
            "cores": settings.common.cores_per_worker,
            "memory": str(settings.common.memory_per_worker) + "GB",
            "processes": settings.common.tasks_per_worker,  # Number of workers to generate == tasks in this construct
            "walltime": settings.cluster.walltime,
            "job_extra": scheduler_opts,
            "env_extra": settings.cluster.task_startup_commands,
            **dask_settings,
        }

        try:
            # Import the dask things we need
            import dask_jobqueue
            from dask.distributed import Client

            cluster_module = cli_utils.import_module(
                "dask_jobqueue", package=_cluster_loaders[settings.cluster.scheduler]
            )
            cluster_class = getattr(cluster_module, _cluster_loaders[settings.cluster.scheduler])
            if dask_jobqueue.__version__ < "0.5.0":
                raise ImportError
        except ImportError:
            raise ImportError("You need`dask-jobqueue >= 0.5.0` to use the `dask` adapter")

        cluster = cluster_class(**dask_construct)

        # Setup up adaption
        # Workers are distributed down to the cores through the sub-divided processes
        # Optimization may be needed
        workers = settings.common.tasks_per_worker * settings.common.max_workers
        if settings.cluster.adaptive == AdaptiveCluster.adaptive:
            cluster.adapt(minimum=0, maximum=workers, interval="10s")
        else:
            cluster.scale(workers)

        queue_client = Client(cluster)

    elif settings.common.adapter == "parsl":

        scheduler_opts = settings.cluster.scheduler_options

        if not settings.cluster.node_exclusivity:
            raise ValueError(
                "For now, QCFractal can only be run with Parsl in node exclusivity. This will be relaxed "
                "in a future release of Parsl and QCFractal"
            )

        # Import helpers
        _provider_loaders = {
            "slurm": "SlurmProvider",
            "pbs": "TorqueProvider",
            "moab": "TorqueProvider",
            "sge": "GridEngineProvider",
            "cobalt": "CobaltProvider",
            "lsf": None,
        }

        if _provider_loaders[settings.cluster.scheduler] is None:
            raise ValueError(f"Parsl does not know how to handle cluster of type {settings.cluster.scheduler}.")

        # Headers
        _provider_headers = {
            "slurm": "#SBATCH",
            "pbs": "#PBS",
            "moab": "#PBS",
            "sge": "#$$",
            "lsf": None,
            "cobalt": "#COBALT",
        }

        # Import the parsl things we need
        try:
            import parsl
            from parsl.config import Config
            from parsl.executors import HighThroughputExecutor
            from parsl.addresses import address_by_hostname

            provider_module = cli_utils.import_module(
                "parsl.providers", package=_provider_loaders[settings.cluster.scheduler]
            )
            provider_class = getattr(provider_module, _provider_loaders[settings.cluster.scheduler])
            provider_header = _provider_headers[settings.cluster.scheduler]
            if parsl.__version__ < "0.9.0":
                raise ImportError
        except ImportError:
            raise ImportError("You need `parsl >=0.9.0` to use the `parsl` adapter")

        if _provider_loaders[settings.cluster.scheduler] == "moab":
            logger.warning(
                "Parsl uses its TorqueProvider for Moab clusters due to the scheduler similarities. "
                "However, if you find a bug with it, please report to the Parsl and QCFractal developers so "
                "it can be fixed on each respective end."
            )

        # Setup the providers

        # Determine the maximum number of blocks
        # TODO (wardlt): Math assumes that user does not set aside a compute node for the adapter
        max_nodes = settings.common.max_workers * settings.common.nodes_per_task
        if settings.common.nodes_per_job > max_nodes:
            raise ValueError("Number of nodes per job is more than the maximum number of nodes used by manager")
        if max_nodes % settings.common.nodes_per_job != 0:
            raise ValueError(
                "Maximum number of nodes (maximum number of workers times nodes per task) "
                "needs to be a multiple of the number of nodes per job"
            )
        if settings.common.nodes_per_job % settings.common.nodes_per_task != 0:
            raise ValueError("Number of nodes per job needs to be a multiple of the number of nodes per task")
        max_blocks = max_nodes // settings.common.nodes_per_job

        # Create one construct to quickly merge dicts with a final check
        common_parsl_provider_construct = {
            "init_blocks": 0,  # Update this at a later time of Parsl
            "max_blocks": max_blocks,
            "walltime": settings.cluster.walltime,
            "scheduler_options": f"{provider_header} " + f"\n{provider_header} ".join(scheduler_opts) + "\n",
            "nodes_per_block": settings.common.nodes_per_job,
            "worker_init": "\n".join(settings.cluster.task_startup_commands),
            **settings.parsl.provider.dict(skip_defaults=True, exclude={"partition", "launcher"}),
        }
        if settings.cluster.scheduler.lower() == "slurm" and "cores_per_node" not in common_parsl_provider_construct:
            common_parsl_provider_construct["cores_per_node"] = settings.common.cores_per_worker
        # TODO: uncomment after Parsl#1416 is resolved
        # if settings.cluster.scheduler.lower() == "slurm" and "mem_per_node" not in common_parsl_provider_construct:
        #    common_parsl_provider_construct["mem_per_node"] = settings.common.memory_per_worker

        if settings.parsl.provider.launcher:
            common_parsl_provider_construct["launcher"] = settings.parsl.provider.launcher.build_launcher()
        if settings.cluster.scheduler == "slurm":
            # The Parsl SLURM constructor has a strange set of arguments
            provider = provider_class(
                settings.parsl.provider.partition,
                exclusive=settings.cluster.node_exclusivity,
                **common_parsl_provider_construct,
            )
        else:
            provider = provider_class(**common_parsl_provider_construct)

        # The executor for Parsl is different for node parallel tasks and shared-memory tasks
        if node_parallel_tasks:
            # Tasks are launched from a single worker on the login node
            # TODO (wardlt): Remove assumption that there is only one Parsl worker running all tasks
            tasks_per_job = settings.common.nodes_per_job // settings.common.nodes_per_task
            logger.info(f"Preparing a HTEx to use node-parallel tasks with {tasks_per_job} workers")
            parsl_executor_construct = {
                "label": "QCFractal_Parsl_{}_Executor".format(settings.cluster.scheduler.title()),
                # Parsl will create one worker process per MPI task. Normally, Parsl prevents having
                #  more processes than cores. However, as each worker will spend most of its time
                #  waiting for the MPI task to complete, we can safely oversubscribe (e.g., more worker
                #  processes than cores), which requires setting "cores_per_worker" to <1
                "cores_per_worker": 1e-6,
                "max_workers": tasks_per_job,
                "provider": provider,
                "address": address_by_hostname(),
                **settings.parsl.executor.dict(skip_defaults=True),
            }
        else:

            parsl_executor_construct = {
                "label": "QCFractal_Parsl_{}_Executor".format(settings.cluster.scheduler.title()),
                "cores_per_worker": cores_per_task,
                "max_workers": settings.common.tasks_per_worker,
                "provider": provider,
                "address": address_by_hostname(),
                **settings.parsl.executor.dict(skip_defaults=True),
            }

        queue_client = Config(
            retries=settings.common.retries, executors=[HighThroughputExecutor(**parsl_executor_construct)]
        )

    else:
        raise KeyError(
            "Unknown adapter type '{}', available options: {}.\n"
            "This code should also be unreachable with pydantic Validation, so if "
            "you see this message, please report it to the QCFractal GitHub".format(
                settings.common.adapter, [getattr(AdapterEnum, v).value for v in AdapterEnum]
            )
        )

    # Build out the manager itself
    # Compute max tasks
    max_concurrent_tasks = settings.common.tasks_per_worker * settings.common.max_workers
    if settings.manager.max_queued_tasks is None:
        # Tasks * jobs * buffer + 1
        max_queued_tasks = ceil(max_concurrent_tasks * 2.00) + 1
    else:
        max_queued_tasks = settings.manager.max_queued_tasks

    # The queue manager is configured differently for node-parallel and single-node tasks
    manager = qcfractal.queue.QueueManager(
        client,
        queue_client,
        max_tasks=max_queued_tasks,
        queue_tag=settings.manager.queue_tag,
        manager_name=settings.manager.manager_name,
        update_frequency=settings.manager.update_frequency,
        cores_per_task=cores_per_task,
        memory_per_task=memory_per_task,
        nodes_per_task=settings.common.nodes_per_task,
        scratch_directory=settings.common.scratch_directory,
        retries=settings.common.retries,
        verbose=settings.common.verbose,
        cores_per_rank=settings.common.cores_per_rank,
        configuration=settings,
    )

    # Set stats correctly since we buffer the max tasks a bit
    manager.statistics.max_concurrent_tasks = max_concurrent_tasks

    # Add exit callbacks
    for cb in exit_callbacks:
        manager.add_exit_callback(cb[0], *cb[1], **cb[2])

    # Either startup the manager or run until complete
    if settings.manager.test:
        success = manager.test(settings.manager.ntests)
        if success is False:
            raise ValueError("Testing was not successful, failing.")
    else:

        for signame in {"SIGHUP", "SIGINT", "SIGTERM"}:

            def stop(*args, **kwargs):
                manager.stop(signame)
                raise KeyboardInterrupt()

            signal.signal(getattr(signal, signame), stop)

        # Blocks until signal
        try:
            manager.start()
        except KeyboardInterrupt:
            pass


if __name__ == "__main__":
    main()



"""
Models for the REST interface
"""
import functools
import re
import warnings
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import Field, constr, root_validator, validator
from qcelemental.util import get_base_docs

from .common_models import KeywordSet, Molecule, ObjectId, ProtoModel, KVStore
from .gridoptimization import GridOptimizationInput
from .records import ResultRecord
from .task_models import PriorityEnum, TaskRecord
from .torsiondrive import TorsionDriveInput

__all__ = [
    "ComputeResponse",
    "rest_model",
    "QueryStr",
    "QueryObjectId",
    "QueryListStr",
    "ResultResponse",
    "CollectionSubresourceGETResponseMeta",
]

### Utility functions

__rest_models = {}


def register_model(name: str, rest: str, body: ProtoModel, response: ProtoModel) -> None:
    """
    Registers a new REST model.

    Parameters
    ----------
    name : str
        A regular expression describing the rest endpoint.
    rest : str
        The REST endpoint type.
    body : ProtoModel
        The REST query body model.
    response : ProtoModel
        The REST query response model.

    """
    rest = rest.upper()

    if (name in __rest_models) and (rest in __rest_models[name]):
        raise KeyError(f"Model name {name} already registered.")

    if name not in __rest_models:
        __rest_models[name] = {}

    __rest_models[name][rest] = (body, response)


@functools.lru_cache(1000, typed=True)
def rest_model(resource: str, rest: str) -> Tuple[ProtoModel, ProtoModel]:
    """
    Acquires a REST Model.

    Parameters
    ----------
    resource : str
        The REST endpoint resource name.
    rest : str
        The REST endpoint type: GET, POST, PUT, DELETE

    Returns
    -------
    Tuple[ProtoModel, ProtoModel]
        The (body, response) models of the REST request.

    """
    rest = rest.upper()
    matches = []
    for model_re in __rest_models.keys():
        if re.fullmatch(model_re, resource):
            try:
                matches.append(__rest_models[model_re][rest])
            except KeyError:
                pass  # Could have different regexes for different endpoint types

    if len(matches) == 0:
        raise KeyError(f"REST Model for endpoint {resource} could not be found.")

    if len(matches) > 1:
        warnings.warn(
            f"Multiple REST models were matched for {rest} request at endpoint {resource}. "
            f"The following models will be used: {matches[0][0]}, {matches[0][1]}.",
            RuntimeWarning,
        )

    return matches[0]


### Generic Types and Common Models

nullstr = constr(regex="null")

QueryStr = Optional[Union[List[str], str]]
QueryInt = Optional[Union[List[int], int]]
QueryObjectId = Optional[Union[List[ObjectId], ObjectId]]
QueryNullObjectId = Optional[Union[List[ObjectId], ObjectId, List[nullstr], nullstr]]
QueryListStr = Optional[List[str]]


class EmptyMeta(ProtoModel):
    """
    There is no metadata accepted, so an empty metadata is sent for completion.
    """


class ResponseMeta(ProtoModel):
    """
    Standard Fractal Server response metadata
    """

    errors: List[Tuple[str, str]] = Field(
        ..., description="A list of error pairs in the form of [(error type, error message), ...]"
    )
    success: bool = Field(
        ...,
        description="Indicates if the passed information was successful in its duties. This is contextual to the "
        "data being passed in.",
    )
    error_description: Union[str, bool] = Field(
        ...,
        description="Details about the error if ``success`` is ``False``, otherwise this is ``False`` in the event "
        "of no errors.",
    )


class ResponseGETMeta(ResponseMeta):
    """
    Standard Fractal Server response metadata for GET/fetch type requests.
    """

    missing: List[str] = Field(..., description="The Id's of the objects which were not found in the database.")
    n_found: int = Field(
        ...,
        description="The number of entries which were already found in the database from the set which was provided.",
    )


class ResponsePOSTMeta(ResponseMeta):
    """
    Standard Fractal Server response metadata for POST/add type requests.
    """

    n_inserted: int = Field(
        ...,
        description="The number of new objects amongst the inputs which did not exist already, and are now in the "
        "database.",
    )
    duplicates: Union[List[str], List[Tuple[str, str]]] = Field(
        ...,
        description="The Ids of the objects which already exist in the database amongst the set which were passed in.",
    )
    validation_errors: List[str] = Field(
        ..., description="All errors with validating submitted objects will be documented here."
    )


class QueryMeta(ProtoModel):
    """
    Standard Fractal Server metadata for Database queries containing pagination information
    """

    limit: Optional[int] = Field(
        None, description="Limit to the number of objects which can be returned with this query."
    )
    skip: int = Field(0, description="The number of records to skip on the query.")


class QueryFilter(ProtoModel):
    """
    Standard Fractal Server metadata for column filtering
    """

    include: QueryListStr = Field(
        None,
        description="Return only these columns. Expert-level object. Only one of include and exclude may be specified.",
    )
    exclude: QueryListStr = Field(
        None,
        description="Return all but these columns. Expert-level object. Only one of include and exclude may be specified.",
    )

    @root_validator
    def check_include_or_exclude(cls, values):
        include = values.get("include")
        exclude = values.get("exclude")
        if (include is not None) and (exclude is not None):
            raise ValueError("Only one of include and exclude may be specified.")
        return values


class QueryMetaFilter(QueryMeta, QueryFilter):
    """
    Fractal Server metadata for Database queries allowing for filtering and pagination
    """


class ComputeResponse(ProtoModel):
    """
    The response model from the Fractal Server when new Compute or Services are added.
    """

    ids: List[Optional[ObjectId]] = Field(..., description="The Id's of the records to be computed.")
    submitted: List[ObjectId] = Field(
        ..., description="The object Ids which were submitted as new entries to the database."
    )
    existing: List[ObjectId] = Field(..., description="The list of object Ids which already existed in the database.")

    def __str__(self) -> str:
        return f"ComputeResponse(nsubmitted={len(self.submitted)} nexisting={len(self.existing)})"

    def __repr__(self) -> str:
        return f"<{self}>"

    def merge(self, other: "ComputeResponse") -> "ComputeResponse":
        """Merges two ComputeResponse objects together. The first takes precedence and order is maintained.

        Parameters
        ----------
        other : ComputeResponse
            The compute response to merge

        Returns
        -------
        ComputeResponse
            The merged compute response
        """
        return ComputeResponse(
            ids=(self.ids + other.ids),
            submitted=(self.submitted + other.submitted),
            existing=(self.existing + other.existing),
        )


common_docs = {
    EmptyMeta: str(get_base_docs(EmptyMeta)),
    ResponseMeta: str(get_base_docs(ResponseMeta)),
    ResponseGETMeta: str(get_base_docs(ResponseGETMeta)),
    ResponsePOSTMeta: str(get_base_docs(ResponsePOSTMeta)),
    QueryMeta: str(get_base_docs(QueryMeta)),
    QueryMetaFilter: str(get_base_docs(QueryMetaFilter)),
    ComputeResponse: str(get_base_docs(ComputeResponse)),
}

### Information


class InformationGETBody(ProtoModel):
    pass


class InformationGETResponse(ProtoModel):
    class Config(ProtoModel.Config):
        extra = "allow"


register_model("information", "GET", InformationGETBody, InformationGETResponse)

### KVStore


class KVStoreGETBody(ProtoModel):
    class Data(ProtoModel):
        id: QueryObjectId = Field(None, description="Id of the Key/Value Storage object to get.")

    meta: EmptyMeta = Field({}, description=common_docs[EmptyMeta])
    data: Data = Field(..., description="Data of the KV Get field: consists of Id of the Key/Value object to fetch.")


class KVStoreGETResponse(ProtoModel):
    meta: ResponseGETMeta = Field(..., description=common_docs[ResponseGETMeta])
    data: Dict[str, KVStore] = Field(..., description="The entries of Key/Value object requested.")


register_model("kvstore", "GET", KVStoreGETBody, KVStoreGETResponse)

### Molecule response


class MoleculeGETBody(ProtoModel):
    class Data(ProtoModel):
        id: QueryObjectId = Field(None, description="Exact Id of the Molecule to fetch from the database.")
        molecule_hash: QueryStr = Field(
            None,
            description="Hash of the Molecule to search for in the database. Can be computed from the Molecule object "
            "directly without direct access to the Database itself.",
        )
        molecular_formula: QueryStr = Field(
            None,
            description="Query is made based on simple molecular formula. This is based on just the formula itself and "
            "contains no connectivity information.",
        )

    meta: QueryMeta = Field(QueryMeta(), description=common_docs[QueryMeta])
    data: Data = Field(
        ...,
        description="Data fields for a Molecule query.",  # Because Data is internal, this may not document sufficiently
    )


class MoleculeGETResponse(ProtoModel):
    meta: ResponseGETMeta = Field(..., description=common_docs[ResponseGETMeta])
    data: List[Molecule] = Field(..., description="The List of Molecule objects found by the query.")


register_model("molecule", "GET", MoleculeGETBody, MoleculeGETResponse)


class MoleculePOSTBody(ProtoModel):
    meta: EmptyMeta = Field({}, description=common_docs[EmptyMeta])
    data: List[Molecule] = Field(..., description="A list of :class:`Molecule` objects to add to the Database.")


class MoleculePOSTResponse(ProtoModel):
    meta: ResponsePOSTMeta = Field(..., description=common_docs[ResponsePOSTMeta])
    data: List[ObjectId] = Field(
        ...,
        description="A list of Id's assigned to the Molecule objects passed in which serves as a unique identifier "
        "in the database. If the Molecule was already in the database, then the Id returned is its "
        "existing Id (entries are not duplicated).",
    )


register_model("molecule", "POST", MoleculePOSTBody, MoleculePOSTResponse)

### Keywords


class KeywordGETBody(ProtoModel):
    class Data(ProtoModel):
        id: QueryObjectId = None
        hash_index: QueryStr = None

    meta: QueryMeta = Field(QueryMeta(), description=common_docs[QueryMeta])
    data: Data = Field(
        ...,
        description="The formal query for a Keyword fetch, contains ``id`` or ``hash_index`` for the object to fetch.",
    )


class KeywordGETResponse(ProtoModel):
    meta: ResponseGETMeta = Field(..., description=common_docs[ResponseGETMeta])
    data: List[KeywordSet] = Field(
        ..., description="The :class:`KeywordSet` found from in the database based on the query."
    )


register_model("keyword", "GET", KeywordGETBody, KeywordGETResponse)


class KeywordPOSTBody(ProtoModel):
    meta: EmptyMeta = Field(
        {}, description="There is no metadata with this, so an empty metadata is sent for completion."
    )
    data: List[KeywordSet] = Field(..., description="The list of :class:`KeywordSet` objects to add to the database.")


class KeywordPOSTResponse(ProtoModel):
    data: List[Optional[ObjectId]] = Field(
        ...,
        description="The Ids assigned to the added :class:`KeywordSet` objects. In the event of duplicates, the Id "
        "will be the one already found in the database.",
    )
    meta: ResponsePOSTMeta = Field(..., description=common_docs[ResponsePOSTMeta])


register_model("keyword", "POST", KeywordPOSTBody, KeywordPOSTResponse)

### Collections


class CollectionGETBody(ProtoModel):
    class Data(ProtoModel):
        collection: str = Field(
            None, description="The specific collection to look up as its identified in the database."
        )
        name: str = Field(None, description="The common name of the collection to look up.")

        @validator("collection")
        def cast_to_lower(cls, v):
            if v:
                v = v.lower()
            return v

    meta: QueryFilter = Field(
        None,
        description="Additional metadata to make with the query. Collections can only have an ``include/exclude`` key in its "
        "meta and therefore does not follow the standard GET metadata model.",
    )
    data: Data = Field(..., description="Information about the Collection to search the database with.")


class CollectionGETResponse(ProtoModel):
    meta: ResponseGETMeta = Field(..., description=common_docs[ResponseGETMeta])
    data: List[Dict[str, Optional[Any]]] = Field(
        ..., description="The Collection objects returned by the server based on the query."
    )

    @validator("data")
    def ensure_collection_name_in_data_get_res(cls, v):
        for col in v:
            if "name" not in col or "collection" not in col:
                raise ValueError("Dicts in 'data' must have both 'collection' and 'name'")
        return v


register_model("collection", "GET", CollectionGETBody, CollectionGETResponse)


class CollectionPOSTBody(ProtoModel):
    class Meta(ProtoModel):
        overwrite: bool = Field(
            False,
            description="The existing Collection in the database will be updated if this is True, otherwise will "
            "remain unmodified if it already exists.",
        )

    class Data(ProtoModel):
        id: str = Field(
            "local",  # Auto blocks overwriting in a socket
            description="The Id of the object to assign in the database. If 'local', then it will not overwrite "
            "existing keys. There should be very little reason to ever touch this.",
        )
        collection: str = Field(
            ..., description="The specific identifier for this Collection as it will appear in database."
        )
        name: str = Field(..., description="The common name of this Collection.")

        class Config(ProtoModel.Config):
            extra = "allow"

        @validator("collection")
        def cast_to_lower(cls, v):
            return v.lower()

    meta: Meta = Field(
        Meta(),
        description="Metadata to specify how the Database should handle adding this Collection if it already exists. "
        "Metadata model for adding Collections can only accept ``overwrite`` as a key to choose to update "
        "existing Collections or not.",
    )
    data: Data = Field(..., description="The data associated with this Collection to add to the database.")


class CollectionPOSTResponse(ProtoModel):
    data: Union[str, None] = Field(
        ...,
        description="The Id of the Collection uniquely pointing to it in the Database. If the Collection was not added "
        "(e.g. ``overwrite=False`` for existing Collection), then a None is returned.",
    )
    meta: ResponsePOSTMeta = Field(..., description=common_docs[ResponsePOSTMeta])


register_model("collection", "POST", CollectionPOSTBody, CollectionPOSTResponse)


class CollectionDELETEBody(ProtoModel):
    meta: EmptyMeta


class CollectionDELETEResponse(ProtoModel):
    meta: ResponseMeta


register_model("collection/[0-9]+", "DELETE", CollectionDELETEBody, CollectionDELETEResponse)

### Collection views


class CollectionSubresourceGETResponseMeta(ResponseMeta):
    """
    Response metadata for collection views functions.
    """

    msgpacked_cols: List[str] = Field(..., description="Names of columns which were serialized to msgpack-ext.")


class CollectionEntryGETBody(ProtoModel):
    class Data(ProtoModel):
        subset: QueryStr = Field(
            None,
            description="Not implemented. " "See qcfractal.interface.collections.dataset_view.DatasetView.get_entries",
        )

    meta: EmptyMeta = Field(EmptyMeta(), description=common_docs[EmptyMeta])
    data: Data = Field(..., description="Information about which entries to return.")


class CollectionEntryGETResponse(ProtoModel):
    meta: CollectionSubresourceGETResponseMeta = Field(
        ..., description=str(get_base_docs(CollectionSubresourceGETResponseMeta))
    )
    data: Optional[bytes] = Field(..., description="Feather-serialized bytes representing a pandas DataFrame.")


register_model("collection/[0-9]+/entry", "GET", CollectionEntryGETBody, CollectionEntryGETResponse)


class CollectionMoleculeGETBody(ProtoModel):
    class Data(ProtoModel):
        indexes: List[int] = Field(
            None,
            description="List of molecule indexes to return (returned by get_entries). "
            "See qcfractal.interface.collections.dataset_view.DatasetView.get_molecules",
        )

    meta: EmptyMeta = Field(EmptyMeta(), description=common_docs[EmptyMeta])
    data: Data = Field(..., description="Information about which molecules to return.")


class CollectionMoleculeGETResponse(ProtoModel):
    meta: CollectionSubresourceGETResponseMeta = Field(
        ..., description=str(get_base_docs(CollectionSubresourceGETResponseMeta))
    )
    data: Optional[bytes] = Field(..., description="Feather-serialized bytes representing a pandas DataFrame.")


register_model("collection/[0-9]+/molecule", "GET", CollectionMoleculeGETBody, CollectionMoleculeGETResponse)


class CollectionValueGETBody(ProtoModel):
    class Data(ProtoModel):
        class QueryData(ProtoModel):
            name: str
            driver: str
            native: bool

        queries: List[QueryData] = Field(
            None,
            description="List of queries to match against values columns. "
            "See qcfractal.interface.collections.dataset_view.DatasetView.get_values",
        )
        subset: QueryStr

    meta: EmptyMeta = Field(EmptyMeta(), description=common_docs[EmptyMeta])
    data: Data = Field(..., description="Information about which values to return.")


class CollectionValueGETResponse(ProtoModel):
    class Data(ProtoModel):
        values: bytes = Field(..., description="Feather-serialized bytes representing a pandas DataFrame.")
        units: Dict[str, str] = Field(..., description="Units of value columns.")

    meta: CollectionSubresourceGETResponseMeta = Field(
        ..., description=str(get_base_docs(CollectionSubresourceGETResponseMeta))
    )
    data: Optional[Data] = Field(..., description="Values and units.")


register_model("collection/[0-9]+/value", "GET", CollectionValueGETBody, CollectionValueGETResponse)


class CollectionListGETBody(ProtoModel):
    class Data(ProtoModel):
        pass

    meta: EmptyMeta = Field(EmptyMeta(), description=common_docs[EmptyMeta])
    data: Data = Field(..., description="Empty for now.")


class CollectionListGETResponse(ProtoModel):
    meta: CollectionSubresourceGETResponseMeta = Field(
        ..., description=str(get_base_docs(CollectionSubresourceGETResponseMeta))
    )
    data: Optional[bytes] = Field(..., description="Feather-serialized bytes representing a pandas DataFrame.")


register_model("collection/[0-9]+/list", "GET", CollectionListGETBody, CollectionListGETResponse)

### Result


class ResultGETBody(ProtoModel):
    class Data(ProtoModel):
        id: QueryObjectId = Field(
            None,
            description="The exact Id to fetch from the database. If this is set as a search condition, there is no "
            "reason to set anything else as this will be unique in the database, if it exists.",
        )
        task_id: QueryObjectId = Field(
            None,
            description="The exact Id of the task which carried out this Result's computation. If this is set as a "
            "search condition, there is no reason to set anything else as this will be unique in the "
            "database, if it exists. See also :class:`TaskRecord`.",
        )

        program: QueryStr = Field(
            None,
            description="Results will be searched to match the quantum chemistry software which carried out the "
            "calculation.",
        )
        molecule: QueryObjectId = Field(
            None, description="Results will be searched to match the Molecule Id which was computed on."
        )
        driver: QueryStr = Field(
            None,
            description="Results will be searched to match what class of computation was done. "
            "See :class:`DriverEnum` for valid choices and more information.",
        )
        method: QueryStr = Field(
            None,
            description="Results will be searched to match the quantum chemistry method executed to compute the value.",
        )
        basis: QueryStr = Field(
            None,
            description="Results will be searched to match specified basis sets which were used to compute the values.",
        )
        keywords: QueryNullObjectId = Field(
            None,
            description="Results will be searched based on which :class:`KeywordSet` was used to run the computation.",
        )

        status: QueryStr = Field(
            "COMPLETE",
            description="Results will be searched based on where they are in the compute pipeline. See the "
            ":class:`RecordStatusEnum` for valid statuses and more information.",
        )

        @validator("keywords", each_item=True, pre=True)
        def validate_keywords(cls, v):
            if v is None:
                v = "null"
            return v

        @validator("basis", each_item=True, pre=True)
        def validate_basis(cls, v):
            if (v is None) or (v == ""):
                v = "null"
            return v

    meta: QueryMetaFilter = Field(QueryMetaFilter(), description=common_docs[QueryMetaFilter])
    data: Data = Field(
        ..., description="The keys with data to search the database on for individual quantum chemistry computations."
    )


class ResultGETResponse(ProtoModel):
    meta: ResponseGETMeta = Field(..., description=common_docs[ResponseGETMeta])
    # Either a record or dict depending if projection
    data: Union[List[ResultRecord], List[Dict[str, Any]]] = Field(
        ...,
        description="Results found from the query. This is a list of :class:`ResultRecord` in most cases, however, "
        "if a projection was specified in the GET request, then a dict is returned with mappings based "
        "on the projection.",
    )

    @validator("data", pre=True)
    def ensure_list_of_dict(cls, v):
        if isinstance(v, dict):
            return [v]
        return v


register_model("result", "GET", ResultGETBody, ResultGETResponse)

### Wavefunction data


class WavefunctionStoreGETBody(ProtoModel):
    class Data(ProtoModel):
        id: ObjectId = Field(None, description="Id of the Wavefunction Key/Value Storage object to get.")

    meta: QueryMetaFilter = Field(QueryMetaFilter(), description=common_docs[QueryMetaFilter])
    data: Data = Field(
        ...,
        description="Data of the Wavefunction Get field: consists of a ObjectId of the Wavefunction object to fetch.",
    )


class WavefunctionStoreGETResponse(ProtoModel):
    meta: ResponseGETMeta = Field(..., description=common_docs[ResponseGETMeta])
    data: Dict[str, Any] = Field(..., description="The entries of the Wavefunction object requested.")


register_model("wavefunctionstore", "GET", WavefunctionStoreGETBody, WavefunctionStoreGETResponse)

### Procedures


class ProcedureGETBody(ProtoModel):
    class Data(ProtoModel):
        id: QueryObjectId = Field(
            None,
            description="The exact Id to fetch from the database. If this is set as a search condition, there is no "
            "reason to set anything else as this will be unique in the database, if it exists.",
        )
        task_id: QueryObjectId = Field(
            None,
            description="The exact Id of a task which is carried out by this Procedure. If this is set as a "
            "search condition, there is no reason to set anything else as this will be unique in the "
            "database, if it exists. See also :class:`TaskRecord`.",
        )

        procedure: QueryStr = Field(None, description="Procedures will be searched based on the name of the procedure.")
        program: QueryStr = Field(
            None,
            description="Procedures will be searched based on the program which is the main manager of the procedure",
        )
        hash_index: QueryStr = Field(
            None,
            description="Procedures will be searched based on a hash of the defined procedure. This is something which "
            "can be generated by the Procedure spec itself and does not require server access to compute. "
            "This should be unique in the database so there should be no reason to set anything else "
            "if this is set as a query.",
        )
        status: QueryStr = Field(
            "COMPLETE",
            description="Procedures will be searched based on where they are in the compute pipeline. See the "
            ":class:`RecordStatusEnum` for valid statuses.",
        )

    meta: QueryMetaFilter = Field(QueryMetaFilter(), description=common_docs[QueryMetaFilter])
    data: Data = Field(..., description="The keys with data to search the database on for Procedures.")


class ProcedureGETResponse(ProtoModel):
    meta: ResponseGETMeta = Field(..., description=common_docs[ResponseGETMeta])
    data: List[Dict[str, Optional[Any]]] = Field(
        ..., description="The list of Procedure specs found based on the query."
    )


register_model("procedure", "GET", ProcedureGETBody, ProcedureGETResponse)

### Task Queue


class TaskQueueGETBody(ProtoModel):
    class Data(ProtoModel):
        id: QueryObjectId = Field(
            None,
            description="The exact Id to fetch from the database. If this is set as a search condition, there is no "
            "reason to set anything else as this will be unique in the database, if it exists.",
        )
        hash_index: QueryStr = Field(
            None,
            description="Tasks will be searched based on a hash of the defined Task. This is something which can "
            "be generated by the Task spec itself and does not require server access to compute. "
            "This should be unique in the database so there should be no reason to set anything else "
            "if this is set as a query.",
        )
        program: QueryStr = Field(
            None, description="Tasks will be searched based on the program responsible for executing this task."
        )
        status: QueryStr = Field(
            None,
            description="Tasks will be search based on where they are in the compute pipeline. See the "
            ":class:`RecordStatusEnum` for valid statuses.",
        )
        base_result: QueryStr = Field(
            None,
            description="The exact Id of the Result which this Task is linked to. If this is set as a "
            "search condition, there is no reason to set anything else as this will be unique in the "
            "database, if it exists. See also :class:`ResultRecord`.",
        )
        tag: QueryStr = Field(None, description="Tasks will be searched based on their associated tag.")
        manager: QueryStr = Field(
            None, description="Tasks will be searched based on the manager responsible for executing the task."
        )

    meta: QueryMetaFilter = Field(QueryMetaFilter(), description=common_docs[QueryMetaFilter])
    data: Data = Field(..., description="The keys with data to search the database on for Tasks.")


class TaskQueueGETResponse(ProtoModel):
    meta: ResponseGETMeta = Field(..., description=common_docs[ResponseGETMeta])
    data: Union[List[TaskRecord], List[Dict[str, Any]]] = Field(
        ...,
        description="Tasks found from the query. This is a list of :class:`TaskRecord` in most cases, however, "
        "if a projection was specified in the GET request, then a dict is returned with mappings based "
        "on the projection.",
    )


register_model("task_queue", "GET", TaskQueueGETBody, TaskQueueGETResponse)


class TaskQueuePOSTBody(ProtoModel):
    class Meta(ProtoModel):
        procedure: str = Field(..., description="Name of the procedure which the Task will execute.")
        program: str = Field(..., description="The program which this Task will execute.")

        tag: Optional[str] = Field(
            None,
            description="Tag to assign to this Task so that Queue Managers can pull only Tasks based on this entry."
            "If no Tag is specified, any Queue Manager can pull this Task.",
        )
        priority: Union[PriorityEnum, None] = Field(None, description=str(PriorityEnum.__doc__))

        class Config(ProtoModel.Config):
            extra = "allow"

        @validator("priority", pre=True)
        def munge_priority(cls, v):
            if isinstance(v, str):
                v = PriorityEnum[v.upper()]
            return v

    meta: Meta = Field(..., description="The additional specification information for the Task to add to the Database.")
    data: List[Union[ObjectId, Molecule]] = Field(
        ...,
        description="The list of either Molecule objects or Molecule Id's (those already in the database) to submit as "
        "part of this Task.",
    )


class TaskQueuePOSTResponse(ProtoModel):

    meta: ResponsePOSTMeta = Field(..., description=common_docs[ResponsePOSTMeta])
    data: ComputeResponse = Field(..., description="Data returned from the server from adding a Task.")


register_model("task_queue", "POST", TaskQueuePOSTBody, TaskQueuePOSTResponse)


class TaskQueuePUTBody(ProtoModel):
    class Data(ProtoModel):
        id: QueryObjectId = Field(
            None,
            description="The exact Id to target in database. If this is set as a search condition, there is no "
            "reason to set anything else as this will be unique in the database, if it exists.",
        )
        base_result: QueryObjectId = Field(  # TODO: Validate this description is correct
            None,
            description="The exact Id of a result which this Task is slated to write to. If this is set as a "
            "search condition, there is no reason to set anything else as this will be unique in the "
            "database, if it exists. See also :class:`ResultRecord`.",
        )
        new_tag: Optional[str] = Field(
            None,
            description="Change the tag of an existing or regenerated task to be this value",
        )
        new_priority: Union[str, int, None] = Field(
            None,
            description="Change the priority of an existing or regenerated task to this value",
        )

    class Meta(ProtoModel):
        operation: str = Field(..., description="The specific action you are taking as part of this update.")

        @validator("operation")
        def cast_to_lower(cls, v):
            return v.lower()

    meta: Meta = Field(..., description="The instructions to pass to the target Task from ``data``.")
    data: Data = Field(..., description="The information which contains the Task target in the database.")


class TaskQueuePUTResponse(ProtoModel):
    class Data(ProtoModel):
        n_updated: int = Field(..., description="The number of tasks which were changed.")

    meta: ResponseMeta = Field(..., description=common_docs[ResponseMeta])
    data: Data = Field(..., description="Information returned from attempting updates of Tasks.")


register_model("task_queue", "PUT", TaskQueuePUTBody, TaskQueuePUTResponse)

### Service Queue


class ServiceQueueGETBody(ProtoModel):
    class Data(ProtoModel):
        id: QueryObjectId = Field(
            None,
            description="The exact Id to fetch from the database. If this is set as a search condition, there is no "
            "reason to set anything else as this will be unique in the database, if it exists.",
        )
        procedure_id: QueryObjectId = Field(  # TODO: Validate this description is correct
            None,
            description="The exact Id of the Procedure this Service is responsible for executing. If this is set as a "
            "search condition, there is no reason to set anything else as this will be unique in the "
            "database, if it exists.",
        )
        hash_index: QueryStr = Field(
            None,
            description="Services are searched based on a hash of the defined Service. This is something which can "
            "be generated by the Service spec itself and does not require server access to compute. "
            "This should be unique in the database so there should be no reason to set anything else "
            "if this is set as a query.",
        )
        status: QueryStr = Field(
            None,
            description="Services are searched based on where they are in the compute pipeline. See the "
            ":class:`RecordStatusEnum` for valid statuses.",
        )

    meta: QueryMeta = Field(QueryMeta(), description=common_docs[QueryMeta])
    data: Data = Field(..., description="The keys with data to search the database on for Services.")


class ServiceQueueGETResponse(ProtoModel):
    meta: ResponseGETMeta = Field(..., description=common_docs[ResponseGETMeta])
    data: List[Dict[str, Optional[Any]]] = Field(
        ..., description="The return of Services found in the database mapping their Ids to the Service spec."
    )


register_model("service_queue", "GET", ServiceQueueGETBody, ServiceQueueGETResponse)


class ServiceQueuePOSTBody(ProtoModel):
    class Meta(ProtoModel):
        tag: Optional[str] = Field(
            None,
            description="Tag to assign to the Tasks this Service will generate so that Queue Managers can pull only "
            "Tasks based on this entry. If no Tag is specified, any Queue Manager can pull this Tasks "
            "created by this Service.",
        )
        priority: Union[str, int, None] = Field(
            None,
            description="Priority given to this Tasks created by this Service. Higher priority will be pulled first.",
        )

    meta: Meta = Field(
        ...,
        description="Metadata information for the Service for the Tag and Priority of Tasks this Service will create.",
    )
    data: List[Union[TorsionDriveInput, GridOptimizationInput]] = Field(
        ..., description="A list the specification for Procedures this Service will manage and generate Tasks for."
    )


class ServiceQueuePOSTResponse(ProtoModel):

    meta: ResponsePOSTMeta = Field(..., description=common_docs[ResponsePOSTMeta])
    data: ComputeResponse = Field(..., description="Data returned from the server from adding a Service.")


register_model("service_queue", "POST", ServiceQueuePOSTBody, ServiceQueuePOSTResponse)


class ServiceQueuePUTBody(ProtoModel):
    class Data(ProtoModel):
        id: QueryObjectId = Field(None, description="The Id of the Service.")
        procedure_id: QueryObjectId = Field(None, description="The Id of the Procedure that the Service is linked to.")

    class Meta(ProtoModel):
        operation: str = Field(..., description="The update action to perform.")

        @validator("operation")
        def cast_to_lower(cls, v):
            return v.lower()

    meta: Meta = Field(..., description="The instructions to pass to the targeted Service.")
    data: Data = Field(..., description="The information which contains the Service target in the database.")


class ServiceQueuePUTResponse(ProtoModel):
    class Data(ProtoModel):
        n_updated: int = Field(..., description="The number of services which were changed.")

    meta: ResponseMeta = Field(..., description=common_docs[ResponseMeta])
    data: Data = Field(..., description="Information returned from attempting updates of Services.")


register_model("service_queue", "PUT", ServiceQueuePUTBody, ServiceQueuePUTResponse)

### Queue Manager


class QueueManagerMeta(ProtoModel):
    """
    Validation and identification Meta information for the Queue Manager's communication with the Fractal Server.
    """

    # Name data
    cluster: str = Field(..., description="The Name of the Cluster the Queue Manager is running on.")
    hostname: str = Field(..., description="Hostname of the machine the Queue Manager is running on.")
    uuid: str = Field(..., description="A UUID assigned to the QueueManager to uniquely identify it.")

    # Username
    username: Optional[str] = Field(None, description="Fractal Username the Manager is being executed under.")

    # Version info
    qcengine_version: str = Field(..., description="Version of QCEngine which the Manager has access to.")
    manager_version: str = Field(
        ..., description="Version of the QueueManager (Fractal) which is getting and returning Jobs."
    )

    # search info
    programs: List[str] = Field(
        ...,
        description="A list of programs which the QueueManager, and thus QCEngine, has access to. Affects which Tasks "
        "the Manager can pull.",
    )
    procedures: List[str] = Field(
        ...,
        description="A list of procedures which the QueueManager has access to. Affects which Tasks "
        "the Manager can pull.",
    )
    tag: QueryStr = Field(
        None,
        description="Optional queue tag to pull Tasks from. If None, tasks are pulled from all tags. "
        "If a list of tags is provided, tasks are pulled in order of tags. (This does not "
        "guarantee tasks will be executed in that order, however.)",
    )

    # Statistics
    total_worker_walltime: Optional[float] = Field(None, description="The total worker walltime in core-hours.")
    total_task_walltime: Optional[float] = Field(None, description="The total task walltime in core-hours.")
    active_tasks: Optional[int] = Field(None, description="The total number of active running tasks.")
    active_cores: Optional[int] = Field(None, description="The total number of active cores.")
    active_memory: Optional[float] = Field(None, description="The total amount of active memory in GB.")


# Add the new QueueManagerMeta to the docs
common_docs[QueueManagerMeta] = str(get_base_docs(QueueManagerMeta))


class QueueManagerGETBody(ProtoModel):
    class Data(ProtoModel):
        limit: int = Field(..., description="Max number of Queue Managers to get from the server.")

    meta: QueueManagerMeta = Field(..., description=common_docs[QueueManagerMeta])
    data: Data = Field(
        ...,
        description="A model of Task request data for the Queue Manager to fetch. Accepts ``limit`` as the maximum "
        "number of tasks to pull.",
    )


class QueueManagerGETResponse(ProtoModel):
    meta: ResponseGETMeta = Field(..., description=common_docs[ResponseGETMeta])
    data: List[Dict[str, Optional[Any]]] = Field(
        ..., description="A list of tasks retrieved from the server to compute."
    )


register_model("queue_manager", "GET", QueueManagerGETBody, QueueManagerGETResponse)


class QueueManagerPOSTBody(ProtoModel):
    meta: QueueManagerMeta = Field(..., description=common_docs[QueueManagerMeta])
    data: Dict[ObjectId, Any] = Field(..., description="A Dictionary of tasks to return to the server.")


class QueueManagerPOSTResponse(ProtoModel):
    meta: ResponsePOSTMeta = Field(..., description=common_docs[ResponsePOSTMeta])
    data: bool = Field(..., description="A True/False return on if the server accepted the returned tasks.")


register_model("queue_manager", "POST", QueueManagerPOSTBody, QueueManagerPOSTResponse)


class QueueManagerPUTBody(ProtoModel):
    class Data(ProtoModel):
        operation: str
        configuration: Optional[Dict[str, Any]] = None

    meta: QueueManagerMeta = Field(..., description=common_docs[QueueManagerMeta])
    data: Data = Field(
        ...,
        description="The update action which the Queue Manager requests the Server take with respect to how the "
        "Queue Manager is tracked.",
    )


class QueueManagerPUTResponse(ProtoModel):
    meta: Dict[str, Any] = Field({}, description=common_docs[EmptyMeta])
    # Order on Union[] is important. Union[bool, Dict[str, int]] -> True if the input dict is not empty since
    # Python can resolve dict -> bool since it passes a `is` test. Will not cast bool -> dict[str, int], so make Dict[]
    # check first
    data: Union[Dict[str, int], bool] = Field(
        ...,
        description="The response from the Server attempting to update the Queue Manager's server-side status. "
        "Response type is a function of the operation made from the PUT request.",
    )


register_model("queue_manager", "PUT", QueueManagerPUTBody, QueueManagerPUTResponse)

## advanced procedures queries


class OptimizationFinalResultBody(ProtoModel):
    class Data(ProtoModel):
        optimization_ids: QueryObjectId = Field(
            None, description="List of optimization procedure Ids to fetch their final results from the database."
        )

    # TODO: not yet supported
    meta: QueryMetaFilter = Field(QueryMetaFilter(), description=common_docs[QueryMetaFilter])
    data: Data = Field(..., description="The keys with data to search the database on for Procedures.")


class OptimizationAllResultBody(ProtoModel):
    class Data(ProtoModel):
        optimization_ids: QueryObjectId = Field(
            None, description="List of optimization procedure Ids to fetch their ALL their results from the database."
        )

    # TODO: not yet supported
    meta: QueryMetaFilter = Field(QueryMetaFilter(), description=common_docs[QueryMetaFilter])
    data: Data = Field(..., description="The keys with data to search the database on for Procedures.")


class OptimizationInitialMoleculeBody(ProtoModel):
    class Data(ProtoModel):
        optimization_ids: QueryObjectId = Field(
            None, description="List of optimization procedure Ids to fetch their initial  molecules from the database."
        )

    # TODO: not yet supported
    meta: QueryMetaFilter = Field(QueryMetaFilter(), description=common_docs[QueryMetaFilter])
    data: Data = Field(..., description="The keys with data to search the database on for Procedures.")


class OptimizationFinalMoleculeBody(ProtoModel):
    class Data(ProtoModel):
        optimization_ids: QueryObjectId = Field(
            None, description="List of optimization procedure Ids to fetch their final molecules from the database."
        )

    # TODO: not yet supported
    meta: QueryMetaFilter = Field(QueryMetaFilter(), description=common_docs[QueryMetaFilter])
    data: Data = Field(..., description="The keys with data to search the database on for Procedures.")


class ResultResponse(ProtoModel):
    meta: ResponseGETMeta = Field(..., description=common_docs[ResponseGETMeta])
    # Either a record or dict depending if projection
    data: Union[Dict[str, ResultRecord], Dict[str, Any]] = Field(
        ..., description="A List of Results found from the query per optimization id."
    )


class ListResultResponse(ProtoModel):
    meta: ResponseGETMeta = Field(..., description=common_docs[ResponseGETMeta])
    # Either a record or dict depending if projection
    data: Union[Dict[str, List[ResultRecord]], Dict[str, Any]] = Field(
        ..., description="A List of Results found from the query per optimization id."
    )


class ListMoleculeResponse(ProtoModel):
    meta: ResponseGETMeta = Field(..., description=common_docs[ResponseGETMeta])
    # Either a record or dict depending if projection
    data: Union[Dict[str, Molecule], Dict[str, Any]] = Field(
        ..., description="A List of Molecules found from the query per optimization id."
    )


register_model(r"optimization/final_result", "GET", OptimizationFinalResultBody, ResultResponse)
register_model(r"optimization/all_results", "GET", OptimizationAllResultBody, ListResultResponse)
register_model(r"optimization/initial_molecule", "GET", OptimizationAllResultBody, ListMoleculeResponse)
register_model(r"optimization/final_molecule", "GET", OptimizationAllResultBody, ListMoleculeResponse)


class ManagerInfoGETBody(ProtoModel):
    class Data(ProtoModel):
        name: QueryStr = Field(None, description="Name(s) of managers to query for.")
        status: QueryStr = Field(
            None,
            description="Managers will be searched based on status. See :class:`ManagerStatusEnum` for valid statuses.",
        )

    meta: QueryMeta = Field(QueryMeta(), description=common_docs[QueryMeta])
    data: Data = Field(..., description="The keys with data to search the database on for Managers.")


class ManagerInfoGETResponse(ProtoModel):
    meta: ResponseGETMeta = Field(..., description=common_docs[ResponseGETMeta])
    data: List[Dict[str, Any]] = Field(..., description="Information about the requested managers")


register_model(r"manager", "GET", ManagerInfoGETBody, ManagerInfoGETResponse)



"""
Main init function for qcfractal
"""

from . import interface

# Handle versioneer
from .extras import get_information

# Handle top level object imports
from .postgres_harness import PostgresHarness, TemporaryPostgres
from .queue import QueueManager
from .server import FractalServer
from .snowflake import FractalSnowflake, FractalSnowflakeHandler

# Import modules
from .storage_sockets import storage_socket_factory

__version__ = get_information("version")
__git_revision__ = get_information("git_revision")
del get_information



from sqlalchemy import JSON, Boolean, Column, ForeignKey, Index, Integer, String
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship
from sqlalchemy.sql.functions import GenericFunction

from qcfractal.storage_sockets.models.sql_base import Base, MsgpackExt

# class json_agg(GenericFunction):
#     type = postgresql.JSON


class json_build_object(GenericFunction):
    type = postgresql.JSON


class CollectionORM(Base):
    """
    A base collection class of precomuted workflows such as datasets, ..

    This is a dynamic document, so it will accept any number of
    extra fields (expandable and uncontrolled schema)
    """

    __tablename__ = "collection"

    id = Column(Integer, primary_key=True)
    collection_type = Column(String)  # for inheritance

    collection = Column(String(100), nullable=False)
    lname = Column(String(100), nullable=False)
    name = Column(String(100), nullable=False)

    tags = Column(JSON)
    tagline = Column(String)
    description = Column(String)

    group = Column(String(100), nullable=False)
    visibility = Column(Boolean, nullable=False)

    view_url_hdf5 = Column(String)
    view_url_plaintext = Column(String)
    view_metadata = Column(JSON)
    view_available = Column(Boolean, nullable=False)

    provenance = Column(JSON)

    extra = Column(JSON)  # extra data related to specific collection type

    def update_relations(self, **kwarg):
        pass

    __table_args__ = (
        Index("ix_collection_lname", "collection", "lname", unique=True),
        Index("ix_collection_type", "collection_type"),
    )

    __mapper_args__ = {"polymorphic_on": "collection_type"}


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class DatasetMixin:
    """
    Mixin class for common Dataset attributes.
    """

    default_benchmark = Column(String)
    default_keywords = Column(JSON)

    default_driver = Column(String)
    default_units = Column(String)
    alias_keywords = Column(JSON)
    default_program = Column(String)

    history_keys = Column(JSON)
    history = Column(JSON)


class ContributedValuesORM(Base):
    """One group of a contibuted values per dataset
    Each dataset can have multiple rows in this table"""

    __tablename__ = "contributed_values"

    collection_id = Column(Integer, ForeignKey("collection.id", ondelete="cascade"), primary_key=True)

    name = Column(String, nullable=False, primary_key=True)
    values = Column(MsgpackExt, nullable=False)
    index = Column(MsgpackExt, nullable=False)
    values_structure = Column(JSON, nullable=False)

    theory_level = Column(JSON, nullable=False)
    units = Column(String, nullable=False)
    theory_level_details = Column(JSON)

    citations = Column(JSON)
    external_url = Column(String)
    doi = Column(String)

    comments = Column(String)


class DatasetEntryORM(Base):
    """Association table for many to many"""

    __tablename__ = "dataset_entry"

    dataset_id = Column(Integer, ForeignKey("dataset.id", ondelete="cascade"), primary_key=True)
    # TODO: check the cascase_delete with molecule
    molecule_id = Column(Integer, ForeignKey("molecule.id"), nullable=False)

    name = Column(String, nullable=False, primary_key=True)
    comment = Column(String)
    local_results = Column(JSON)


class DatasetORM(CollectionORM, DatasetMixin):
    """
    The Dataset class for homogeneous computations on many molecules.
    """

    __tablename__ = "dataset"

    id = Column(Integer, ForeignKey("collection.id", ondelete="CASCADE"), primary_key=True)

    contributed_values_obj = relationship(ContributedValuesORM, lazy="selectin", cascade="all, delete-orphan")

    records_obj = relationship(
        DatasetEntryORM, lazy="selectin", cascade="all, delete-orphan", backref="dataset"  # lazy='noload',
    )

    @hybrid_property
    def contributed_values(self):
        return self._contributed_values(self.contributed_values_obj)

    @staticmethod
    def _contributed_values(contributed_values_obj):
        if not contributed_values_obj:
            return {}

        if not isinstance(contributed_values_obj, list):
            contributed_values_obj = [contributed_values_obj]
        ret = {}
        try:
            for obj in contributed_values_obj:
                ret[obj.name.lower()] = obj.to_dict(exclude=["collection_id"])
        except Exception as err:
            pass

        return ret

    @contributed_values.setter
    def contributed_values(self, dict_values):
        return dict_values

    @hybrid_property
    def records(self):
        """calculated property when accessed, not saved in the DB
        A view of the many to many relation"""

        return self._records(self.records_obj)

    @staticmethod
    def _records(records_obj):

        if not records_obj:
            return []

        if not isinstance(records_obj, list):
            records_obj = [records_obj]

        ret = []
        try:
            for rec in records_obj:
                ret.append(rec.to_dict(exclude=["dataset_id"]))
        except Exception as err:
            # raises exception of first access!!
            pass

        return ret

    @records.setter
    def records(self, dict_values):
        return dict_values

    def update_relations(self, records=None, contributed_values=None, **kwarg):

        self.records_obj = []
        records = [] if not records else records
        for rec_dict in records:
            rec = DatasetEntryORM(dataset_id=int(self.id), **rec_dict)
            self.records_obj.append(rec)

        self.contributed_values_obj = []
        contributed_values = {} if not contributed_values else contributed_values
        for key, rec_dict in contributed_values.items():
            rec = ContributedValuesORM(collection_id=int(self.id), **rec_dict)
            self.contributed_values_obj.append(rec)

    __table_args__ = (
        # Index('ix_results_molecule', 'molecule'),  # b-tree index
        # UniqueConstraint("program", "driver", "method", "basis", "keywords", "molecule", name='uix_results_keys'),
    )

    __mapper_args__ = {
        "polymorphic_identity": "dataset",
        # to have separate select when querying CollectionORM
        "polymorphic_load": "selectin",
    }


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class ReactionDatasetEntryORM(Base):
    """Association table for many to many"""

    __tablename__ = "reaction_dataset_entry"

    reaction_dataset_id = Column(Integer, ForeignKey("reaction_dataset.id", ondelete="cascade"), primary_key=True)

    attributes = Column(JSON)
    name = Column(String, nullable=False, primary_key=True)
    reaction_results = Column(JSON)
    stoichiometry = Column(JSON)
    extras = Column(JSON)


class ReactionDatasetORM(CollectionORM, DatasetMixin):
    """
    Reaction Dataset
    """

    __tablename__ = "reaction_dataset"

    id = Column(Integer, ForeignKey("collection.id", ondelete="CASCADE"), primary_key=True)

    ds_type = Column(String)

    records_obj = relationship(
        ReactionDatasetEntryORM, lazy="selectin", cascade="all, delete-orphan", backref="reaction_dataset"
    )

    contributed_values_obj = relationship(ContributedValuesORM, lazy="selectin", cascade="all, delete-orphan")

    @hybrid_property
    def contributed_values(self):
        return self._contributed_values(self.contributed_values_obj)

    @staticmethod
    def _contributed_values(contributed_values_obj):
        return DatasetORM._contributed_values(contributed_values_obj)

    @contributed_values.setter
    def contributed_values(self, dict_values):
        return dict_values

    def update_relations(self, records=None, contributed_values=None, **kwarg):

        self.records_obj = []
        records = records or []
        for rec_dict in records:
            rec = ReactionDatasetEntryORM(reaction_dataset_id=int(self.id), **rec_dict)
            self.records_obj.append(rec)

        self.contributed_values_obj = []
        contributed_values = {} if not contributed_values else contributed_values
        for key, rec_dict in contributed_values.items():
            rec = ContributedValuesORM(collection_id=int(self.id), **rec_dict)
            self.contributed_values_obj.append(rec)

    @hybrid_property
    def records(self):
        """calculated property when accessed, not saved in the DB
        A view of the many to many relation"""

        return self._records(self.records_obj)

    @staticmethod
    def _records(records_obj):

        if not records_obj:
            return []

        if not isinstance(records_obj, list):
            records_obj = [records_obj]

        ret = []
        try:
            for rec in records_obj:
                ret.append(rec.to_dict(exclude=["reaction_dataset_id"]))
        except Exception as err:
            # raises exception of first access!!
            pass
        return ret

    @records.setter
    def records(self, dict_values):
        return dict_values

    __table_args__ = (
        # Index('ix_results_molecule', 'molecule'),  # b-tree index
        # UniqueConstraint("program", "driver", "method", "basis", "keywords", "molecule", name='uix_results_keys'),
    )

    __mapper_args__ = {
        "polymorphic_identity": "reactiondataset",
        # to have separate select when querying CollectionORM
        "polymorphic_load": "selectin",
    }


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



import pytest
from pydantic import BaseModel, ValidationError

from ..common_models import ObjectId


class OIdModel(BaseModel):
    id: ObjectId


@pytest.mark.parametrize(
    "oid", ["5c754f049642c7c861d67de5", "000000000000000000000000", "0123456789abcdef01234567", "123", 123]
)
def test_objid_check(oid):
    OIdModel(id=oid)


@pytest.mark.parametrize("oid", ["", "0123456789abcdef0123456z"])
def test_objid_wrong(oid):
    with pytest.raises(ValidationError):
        OIdModel(id=oid)



"""service ordering

Revision ID: 6adeb5d3032e
Revises: 25f32879e0fb
Create Date: 2019-07-09 12:02:39.164203

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "6adeb5d3032e"
down_revision = "d28e52d86633"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("service_queue", sa.Column("created_on", sa.DateTime(), nullable=True))
    op.add_column("service_queue", sa.Column("modified_on", sa.DateTime(), nullable=True))
    op.add_column("service_queue", sa.Column("priority", sa.Integer(), nullable=True))
    op.create_index("ix_service_queue_modified_on", "service_queue", ["modified_on"], unique=False)
    op.create_index("ix_service_queue_priority", "service_queue", ["priority"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_service_queue_priority", table_name="service_queue")
    op.drop_index("ix_service_queue_modified_on", table_name="service_queue")
    op.drop_column("service_queue", "priority")
    op.drop_column("service_queue", "modified_on")
    op.drop_column("service_queue", "created_on")
    # ### end Alembic commands ###



import logging
import sqlalchemy as sa
import tqdm
import numpy as np
from alembic import op
from sqlalchemy.dialects.postgresql import BYTEA
from sqlalchemy.sql.expression import func

from qcelemental.util import msgpackext_dumps, msgpackext_loads
from qcelemental.testing import compare_recursive

logger = logging.getLogger("alembic")

old_type = sa.JSON
new_type = BYTEA


def _get_colnames(columns):
    pairs = {(x, x + "_") for x in columns}
    old = [x[0] for x in pairs]
    new = [x[1] for x in pairs]
    return (pairs, old, new)


def _intermediate_table(table_name, columns, read_columns=None):

    column_pairs, old_names, new_names = _get_colnames(columns)
    table_data = [table_name, sa.MetaData(), sa.Column("id", sa.Integer, primary_key=True)]
    table_data.extend([sa.Column(x, old_type) for x in old_names])
    table_data.extend([sa.Column(x, new_type) for x in new_names])

    cols = old_names + new_names

    if read_columns:
        table_data.extend([sa.Column(k, v) for k, v in read_columns.items()])

        cols.extend(list(read_columns))

    table = sa.Table(*table_data)
    return table, cols


def json_to_msgpack_table(table_name, block_size, update_columns, transformer, read_columns=None):

    if read_columns is None:
        read_columns = {}

    update_columns = list(update_columns)

    logger.info(f"Converting {table_name} from JSON to msgpack.")
    logger.info(f"Columns: {update_columns}")
    column_pairs, old_names, new_names = _get_colnames(update_columns)

    # Schema migration: add all the new columns.
    for col_old, col_new in column_pairs:
        op.add_column(table_name, sa.Column(col_new, new_type, nullable=True))

    # Declare a view of the table
    table, cols = _intermediate_table(table_name, update_columns, read_columns=read_columns)

    connection = op.get_bind()

    num_records = connection.execute(f"select count(*) from {table_name}").scalar()

    read_names = ["id"] + old_names + list(read_columns)
    read_columns = [getattr(table.c, x) for x in read_names]

    logger.info("Converting data, this may take some time...")
    for block in tqdm.tqdm(range(0, num_records, block_size)):

        # Pull chunk to migrate
        data = connection.execute(
            sa.select([*read_columns], order_by=table.c.id.asc(), offset=block, limit=block_size)
        ).fetchall()

        # Convert chunk to msgpack
        for values in data:
            data = {k: v for k, v in zip(read_names, values)}
            row = transformer(data)

            connection.execute(table.update().where(table.c.id == data["id"]).values(**row))

        connection.execute("commit;")


def json_to_msgpack_table_dropcols(table_name, block_size, update_columns):

    column_pairs, old_names, new_names = _get_colnames(update_columns)
    for col in new_names:
        op.drop_column(table_name, col)


def json_to_msgpack_table_altercolumns(table_name, update_columns, nullable_true=None):

    if nullable_true is None:
        nullable_true = set()

    connection = op.get_bind()
    table, cols = _intermediate_table(table_name, update_columns)

    column_pairs, old_names, new_names = _get_colnames(update_columns)
    num_records = connection.execute(f"select count(*) from {table_name}").scalar()

    old_columns = [getattr(table.c, x) for x in old_names]
    new_columns = [getattr(table.c, x) for x in new_names]

    logger.info(f"Checking converted columns...")
    # Pull chunk to migrate
    data = connection.execute(sa.select([table.c.id, *old_columns, *new_columns], order_by=table.c.id.asc())).fetchall()
    # ], limit=100, order_by=func.random())).fetchall()

    col_names = ["id"] + old_names + new_names
    for values in data:
        row = {k: v for k, v in zip(col_names, values)}
        # print(row["id"])
        # print(row.keys())
        # for k, v in row.items():
        #     print(k, v)

        for name in old_names:
            comp_data = msgpackext_loads(row[name + "_"])
            # try:
            #     assert compare_recursive(comp_data, row[name])
            # except AssertionError:
            #     assert compare_recursive(comp_data.ravel(), row[name], quiet=True)

            # try:
            #     print(name, comp_data.dtype, comp_data)
            # except:
            #     print(name, comp_data[0].dtype, comp_data)
            #     pass
    # raise Exception()
    logger.info(f"Dropping old columns and renaming new.")
    # Drop old tables and swamp new ones in.
    for old_name, new_name in column_pairs:
        nullable = False
        if old_name in nullable_true:
            nullable = True

        op.drop_column(table_name, old_name)
        op.alter_column(table_name, new_name, new_column_name=old_name, nullable=nullable)
